﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Xml;
using Celeste;
using Celeste.Mod;
using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.Utils;
using YamlDotNet.Serialization;

namespace TimeTrials
{

    public class TimeTrialSettings : EverestModuleSettings
    {
        public bool Enabled { get; set; } = true;
    }

    public class TimeTrialModule : EverestModule
    {
        /// <summary>
        /// SID, [Platinum, Gold, Silver, Bronze] - in seconds
        /// </summary>
        public static Dictionary<string, Tuple<int, int, int, int>> ASideTimes = new Dictionary<string, Tuple<int, int, int, int>>()
        {
            { "Celeste/1-ForsakenCity", new Tuple<int, int, int, int>(60, 70, 80, 100)},
            { "Celeste/2-OldSite", new Tuple<int, int, int, int>(105, 135, 175, 200)},
            { "Celeste/3-CelestialResort", new Tuple<int, int, int, int>(260, 300, 350, 400)},
            { "Celeste/4-GoldenRidge", new Tuple<int, int, int, int>(160, 200, 250, 300)},
            { "Celeste/5-MirrorTemple", new Tuple<int, int, int, int>(260, 300, 350, 400)},
            { "Celeste/6-Reflection", new Tuple<int, int, int, int>(360, 400, 450, 500)},
            { "Celeste/7-Summit", new Tuple<int, int, int, int>(520, 700, 800, 900)},
            { "Celeste/9-Core", new Tuple<int, int, int, int>(250, 300, 350, 500)},
            { "Celeste/LostLevels", new Tuple<int, int, int, int>(1260, 1500, 1800, 2400)}
        };

        /// <summary>
        /// SID, [Platinum, Gold, Silver, Bronze] - in seconds
        /// </summary>
        public static Dictionary<string, Tuple<int, int, int, int>> BSideTimes = new Dictionary<string, Tuple<int, int, int, int>>()
        {
            { "Celeste/1-ForsakenCity", new Tuple<int, int, int, int>(120, 150, 160, 240)},
            { "Celeste/2-OldSite", new Tuple<int, int, int, int>(125, 150, 160, 240)},
            { "Celeste/3-CelestialResort", new Tuple<int, int, int, int>(150, 180, 260, 400)},
            { "Celeste/4-GoldenRidge", new Tuple<int, int, int, int>(160, 200, 250, 300)},
            { "Celeste/5-MirrorTemple", new Tuple<int, int, int, int>(190, 250, 350, 425)},
            { "Celeste/6-Reflection", new Tuple<int, int, int, int>(280, 400, 500, 600)},
            { "Celeste/7-Summit", new Tuple<int, int, int, int>(420, 500, 800, 900)},
            { "Celeste/9-Core", new Tuple<int, int, int, int>(230, 300, 350, 500)}
        };

        /// <summary>
        /// SID, [Platinum, Gold, Silver, Bronze] - in seconds
        /// </summary>
        public static Dictionary<string, Tuple<int, int, int, int>> CSideTimes = new Dictionary<string, Tuple<int, int, int, int>>()
        {
            { "Celeste/1-ForsakenCity", new Tuple<int, int, int, int>(20, 30, 35, 40)},
            { "Celeste/2-OldSite", new Tuple<int, int, int, int>(25, 30, 40, 45)},
            { "Celeste/3-CelestialResort", new Tuple<int, int, int, int>(25, 30, 35, 50)},
            { "Celeste/4-GoldenRidge", new Tuple<int, int, int, int>(30, 40, 45, 50)},
            { "Celeste/5-MirrorTemple", new Tuple<int, int, int, int>(25, 30, 35, 40)},
            { "Celeste/6-Reflection", new Tuple<int, int, int, int>(35, 40, 45, 50)},
            { "Celeste/7-Summit", new Tuple<int, int, int, int>(45, 60, 65, 80)},
            { "Celeste/9-Core", new Tuple<int, int, int, int>(60, 70, 80, 90)}
        };

        public static Dictionary<string, Tuple<int, int, int>> CachedTimes = new Dictionary<string, Tuple<int, int, int>>()
        {

        };

        public static void ReadDataXml(string path)
        {
            Logger.Log("Time Trials", $"Reading {path}");
            XmlElement element = Calc.LoadXML(path)["TimeTrials"];
            foreach (XmlElement map in element.ChildNodes)
            {
                string SID = map.Attributes["SID"].InnerText;
                //SID = SID.Replace("/", "s2F");
                if (!ASideTimes.ContainsKey(SID))
                {
                    Logger.Log("Time Trials", $"SID: {SID}");
                    // Now let's read all the sides:
                    foreach (XmlElement side in map.ChildNodes)
                    {
                        int bronze = int.Parse(side.ChildNodes[0].InnerText);
                        int silver = int.Parse(side.ChildNodes[1].InnerText);
                        int gold = int.Parse(side.ChildNodes[2].InnerText);
                        int platinum = int.Parse(side.ChildNodes[3].InnerText);

                        Tuple<int, int, int, int> tuple = new Tuple<int, int, int, int>(platinum, gold, silver, bronze);
                        Logger.Log("Time Trials", $"SIDE: {side.Name}");
                        Logger.Log("Time Trials", $"Times: {tuple.ToString()}");
                        switch (side.Name)
                        {
                            case "Normal":
                                ASideTimes.Add(SID, tuple);
                                break;
                            case "BSide":
                                BSideTimes.Add(SID, tuple);
                                break;
                            case "CSide":
                                CSideTimes.Add(SID, tuple);
                                break;
                        }
                    }
                } else
                {
                    Logger.Log("Time Trials", $"[WARN] Found duplicate map SID {SID}!");
                }
                
            }
            
        }

        public static void ReadAllXmls()
        {
            Dictionary<string, ModAsset> map = Everest.Content.Map;
            lock (map)
            {
                foreach (ModAsset modAsset in Everest.Content.Map.Values.ToArray())
                {
                    if ((modAsset?.Type == typeof(AssetTypeXml)) && modAsset.PathVirtual.StartsWith("TimeTrials/"))
                    {
                        Logger.Log("Time Trials", $"Found xml: {modAsset.PathVirtual} => {modAsset.GetCachedPath()}");
                        ReadDataXml(modAsset.GetCachedPath());
                    }
                }
            }

        }

        // Only one alive module instance can exist at any given time.
        public static TimeTrialModule Instance;

        public TimeTrialModule()
        {
            Instance = this;
        }
        // no save data needed
        public override Type SaveDataType => null;

        public override Type SettingsType => typeof(TimeTrialSettings);

        public static TimeTrialSettings Settings => (TimeTrialSettings)Instance._Settings;

        public override void LoadContent(bool firstLoad)
        {
            // Create xml if it doesn't exist yet
            LoadTime("ada", AreaMode.Normal);
            bg = GFX.Gui["strawberryCountBG"];
            ReadAllXmls();
        }


        public int LoadTime(string SID, AreaMode side)
        {
            XmlDocument doc;
            // xml names can't have /
            SID = SID.Replace("/", "s2F");
            if (CachedTimes.ContainsKey(SID))
            {
                switch (side)
                {
                    case AreaMode.Normal:
                        return CachedTimes[SID].Item1;
                    case AreaMode.BSide:
                        return CachedTimes[SID].Item2;
                    case AreaMode.CSide:
                        return CachedTimes[SID].Item3;
                }
            } else
            {
                string path = Path.Combine(Everest.PathSettings, "timeTrialSave.xml");
                if (File.Exists(path))
                {
                    doc = Calc.LoadXML(path);
                }
                else
                {
                    doc = new XmlDocument();
                    XmlElement tims = doc.CreateElement("times");
                    doc.AppendChild(tims);
                    doc.Save(path);
                }

                XmlElement times = doc["times"];
                if (times.HasChild(SID))
                {
                    XmlElement child = times[SID];
                    if (child.HasChild(side.ToString()))
                    {
                        return int.Parse(child[side.ToString()].InnerText);
                    }
                    else
                    {
                        child.CreateChild(side.ToString()).InnerText = "-1";
                    }
                }
            }
            return -1;
        }

        public void SaveTime(string SID, AreaMode side, int Rank)
        {
            // xml names can't have /
            SID = SID.Replace("/", "s2F");
            string path = Path.Combine(Everest.PathSettings, "timeTrialSave.xml");
            XmlDocument doc = Calc.LoadXML(path);
            XmlElement times = doc["times"];
            if (times.HasChild(SID))
            {
                XmlElement child = times[SID];
                if (child.HasChild(side.ToString()))
                {
                    child[side.ToString()].InnerText = Rank.ToString();
                } else
                {
                    child.CreateChild(side.ToString()).InnerText = Rank.ToString();
                }
            } else
            {
                times.CreateChild(SID);//.InnerText = Rank.ToString();
                times[SID].CreateChild(side.ToString()).InnerText = Rank.ToString();
            }
            doc.Save(path);
            Tuple<int, int, int> tuple;
            int i1 = 0;
            int i2 = 0;
            int i3 = 0;
            if (CachedTimes.ContainsKey(SID))
            {
                Tuple<int, int, int> old = CachedTimes[SID];
                i1 = old.Item1;
                i2 = old.Item2;
                i3 = old.Item3;
            }
            switch (side)
            {
                case AreaMode.Normal:
                    i1 = Rank;
                    break;
                case AreaMode.BSide:
                    i2 = Rank;
                    break;
                case AreaMode.CSide:
                    i3 = Rank;
                    break;
            }
            if (CachedTimes.ContainsKey(SID))
            {
                CachedTimes[SID] = new Tuple<int, int, int>(i1, i2, i3);
            } else
            {
                CachedTimes.Add(SID, new Tuple<int, int, int>(i1, i2, i3));
            } 
        }

        const int BerryTimeBonus = 5;
        static int BerryTime {
            get
            {
                if ((Engine.Scene as Level) != null)
                    return (Engine.Scene as Level).Session.GetCounter("timeTrialsBerryCount");
                return 0;
            }
            set
            {
                if ((Engine.Scene as Level) != null)
                    (Engine.Scene as Level).Session.SetCounter("timeTrialsBerryCount", value);
            }
        }
        // Set up any hooks, event handlers and your mod in general here.
        // Load runs before Celeste itself has initialized properly.
        public override void Load()
        {
            On.Celeste.SpeedrunTimerDisplay.DrawTime += SpeedrunTimerDisplay_DrawTime;
            On.Celeste.SpeedrunTimerDisplay.CalculateBaseSizes += SpeedrunTimerDisplay_CalculateBaseSizes;
            On.Celeste.Level.RegisterAreaComplete += Level_RegisterAreaComplete;

            On.Celeste.Strawberry.OnCollect += Strawberry_OnCollect;

            On.Celeste.OuiChapterPanel.Render += OuiChapterPanel_Render;
        }

        private void OuiChapterPanel_Render(On.Celeste.OuiChapterPanel.orig_Render orig, OuiChapterPanel self)
        {
            orig(self);
            MTexture mtexture = GFX.Gui["time" + RankToString(LoadTime(self.Data.SID, self.Area.Mode))];
            mtexture.Draw(self.Position + new Vector2(60f, 224f), Vector2.Zero, Color.White, 3f);
        }

        private void Strawberry_OnCollect(On.Celeste.Strawberry.orig_OnCollect orig, Strawberry self)
        {
            orig(self);
            BerryTime = BerryTime + BerryTimeBonus;
        }

        private void Level_RegisterAreaComplete(On.Celeste.Level.orig_RegisterAreaComplete orig, Level self)
        {
            if (self.Session.StartedFromBeginning)
            {
                // just finished level, save if we got a new best rank
                string SID = self.Session.Area.SID;
                AreaMode side = self.Session.Area.Mode;
                TimeSpan span = TimeSpan.FromTicks(self.Session.Time);
                int time = (int)span.TotalSeconds - BerryTime;
                int PlatinumTime;
                int GoldTime;
                int SilverTime;
                int BronzeTime;
                // A side
                if ((Engine.Scene as Level).Session.Area.Mode == AreaMode.Normal)
                {
                    if (ASideTimes.ContainsKey((Engine.Scene as Level).Session.Area.SID))
                    {
                        Tuple<int, int, int, int> tuple = ASideTimes[(Engine.Scene as Level).Session.Area.SID];
                        PlatinumTime = tuple.Item1;
                        GoldTime = tuple.Item2;
                        SilverTime = tuple.Item3;
                        BronzeTime = tuple.Item4;
                    }
                    else
                    {
                        orig(self);
                        return;
                    }
                }
                else
                // B side
                if ((Engine.Scene as Level).Session.Area.Mode == AreaMode.BSide)
                {
                    if (BSideTimes.ContainsKey((Engine.Scene as Level).Session.Area.SID))
                    {
                        Tuple<int, int, int, int> tuple = BSideTimes[(Engine.Scene as Level).Session.Area.SID];
                        PlatinumTime = tuple.Item1;
                        GoldTime = tuple.Item2;
                        SilverTime = tuple.Item3;
                        BronzeTime = tuple.Item4;
                    }
                    else
                    {
                        orig(self);
                        return;
                    }
                }
                else // C side
                {
                    if (CSideTimes.ContainsKey((Engine.Scene as Level).Session.Area.SID))
                    {
                        Tuple<int, int, int, int> tuple = CSideTimes[(Engine.Scene as Level).Session.Area.SID];
                        PlatinumTime = tuple.Item1;
                        GoldTime = tuple.Item2;
                        SilverTime = tuple.Item3;
                        BronzeTime = tuple.Item4;
                    }
                    else
                    {
                        orig(self);
                        return;
                    }
                }
                // Get the rank
                int Rank = -1;
                if (time < PlatinumTime)
                {
                    Rank = 3;
                }
                else if (time < GoldTime)
                {
                    Rank = 2;
                }
                else if (time < SilverTime)
                {
                    Rank = 1;
                }
                else if (time < BronzeTime)
                {
                    Rank = 0;
                }
                if (Rank > LoadTime(SID, side))
                {
                    SaveTime(SID, side, Rank);
                    (Engine.Scene as Level).Session.SetFlag("timeTrialsNewBest", true);
                }
            }
            orig(self);
            
        }

        private void SpeedrunTimerDisplay_CalculateBaseSizes(On.Celeste.SpeedrunTimerDisplay.orig_CalculateBaseSizes orig)
        {
            orig();
            // f you devs
            FieldInfo nw = typeof(SpeedrunTimerDisplay).GetField("numberWidth", BindingFlags.NonPublic | BindingFlags.Static);
            FieldInfo sw = typeof(SpeedrunTimerDisplay).GetField("spacerWidth", BindingFlags.NonPublic | BindingFlags.Static);
            numberWidth = (float)nw.GetValue(null);
            spacerWidth = (float)sw.GetValue(null);
        }

        float numberWidth;
        float spacerWidth;

        public List<Color> Colors = new List<Color>()
        {
            // Platinum
            Calc.HexToColor("a0b2c6"),
            // Gold
            Calc.HexToColor("ffd700"),
            // Silver
            Calc.HexToColor("aaa9ad"),
            // Bronze
            Calc.HexToColor("cd7f32")
        };

        static MTexture bg;

        string FormatTime(int seconds)
        {
            int left = seconds % 60;
            string time = (seconds / 60) + ":" + (left < 10 ? "0" : "") + left;
            return time;
        }

        private void SpeedrunTimerDisplay_DrawTime(On.Celeste.SpeedrunTimerDisplay.orig_DrawTime orig, Vector2 position, string timeString, float scale, bool valid, bool finished, bool bestTime, float alpha)
        {
            if (Settings.Enabled && (Engine.Scene as Level) != null)
            {
                int PlatinumTime;
                int GoldTime;
                int SilverTime;
                int BronzeTime;
                // A side
                if ((Engine.Scene as Level).Session.Area.Mode == AreaMode.Normal)
                {
                    if (ASideTimes.ContainsKey((Engine.Scene as Level).Session.Area.SID))
                    {
                        Tuple<int, int, int, int> tuple = ASideTimes[(Engine.Scene as Level).Session.Area.SID];
                        PlatinumTime = tuple.Item1;
                        GoldTime = tuple.Item2;
                        SilverTime = tuple.Item3;
                        BronzeTime = tuple.Item4;
                    } else
                    {
                        orig(position, timeString, scale, valid, finished, bestTime, alpha);
                        return;
                    }
                } else
                // B side
                if ((Engine.Scene as Level).Session.Area.Mode == AreaMode.BSide)
                {
                    if (BSideTimes.ContainsKey((Engine.Scene as Level).Session.Area.SID))
                    {
                        Tuple<int, int, int, int> tuple = BSideTimes[(Engine.Scene as Level).Session.Area.SID];
                        PlatinumTime = tuple.Item1;
                        GoldTime = tuple.Item2;
                        SilverTime = tuple.Item3;
                        BronzeTime = tuple.Item4;
                    }
                    else
                    {
                        orig(position, timeString, scale, valid, finished, bestTime, alpha);
                        return;
                    }
                }
                else // C side
                {
                    if (CSideTimes.ContainsKey((Engine.Scene as Level).Session.Area.SID))
                    {
                        Tuple<int, int, int, int> tuple = CSideTimes[(Engine.Scene as Level).Session.Area.SID];
                        PlatinumTime = tuple.Item1;
                        GoldTime = tuple.Item2;
                        SilverTime = tuple.Item3;
                        BronzeTime = tuple.Item4;
                    }
                    else
                    {
                        orig(position, timeString, scale, valid, finished, bestTime, alpha);
                        return;
                    }
                }
                // Copied
                PixelFont font = Dialog.Languages["english"].Font;
                float fontFaceSize = Dialog.Languages["english"].FontFaceSize;
                float num = scale;
                float num2 = position.X;
                float num3 = position.Y;
                Color color = Color.White * alpha;
                Color color2 = Color.LightGray * alpha;
                int curRank = -1;
                int targetTime = -1;
                if (!valid)
                {
                    color = Calc.HexToColor("918988") * alpha;
                    color2 = Calc.HexToColor("7a6f6d") * alpha;
                }
                //else if (bestTime)
                //{
                //    color = Calc.HexToColor("fad768") * alpha;
                //    color2 = Calc.HexToColor("cfa727") * alpha;
                //} else
                else
                {
                    // Get the time:
                    Level level = Engine.Scene as Level;
                    Session session = level.Session;
                    TimeSpan span = TimeSpan.FromTicks(session.Time);
                    int time = (int)span.TotalSeconds - BerryTime;
                    // Get the color

                    if (time < PlatinumTime)
                    {
                        color = Colors[0] * alpha;
                        color2 = color;
                        curRank = 3;
                        targetTime = PlatinumTime;
                    }
                    else if (time < GoldTime)
                    {
                        color = Colors[1] * alpha;
                        color2 = color;
                        curRank = 2;
                        targetTime = GoldTime;
                    }
                    else if (time < SilverTime)
                    {
                        color = Colors[2] * alpha;
                        color2 = color;
                        curRank = 1;
                        targetTime = SilverTime;
                    }
                    else if (time < BronzeTime)
                    {
                        color = Colors[3] * alpha;
                        color2 = color;
                        curRank = 0;
                        targetTime = BronzeTime;
                    }
                    else
                    {
                        color = Color.White * alpha;
                        color2 = Color.LightGray * alpha;
                        curRank = -1;
                    }
                }
                //else if (finished)
                //{
                //    color = Calc.HexToColor("6ded87") * alpha;
                //    color2 = Calc.HexToColor("43d14c") * alpha;
                //}
                foreach (char c in timeString)
                {
                    bool flag2 = c == '.';
                    if (flag2)
                    {
                        num = scale * 0.7f;
                        num3 -= 5f * scale;
                    }
                    Color color3 = (c == ':' || c == '.' || num < scale) ? color2 : color;
                    float num4 = (((c == ':' || c == '.') ? spacerWidth : numberWidth) + 4f) * num;
                    font.DrawOutline(fontFaceSize, c.ToString(), new Vector2(num2 + num4 / 2f, num3), new Vector2(0.5f, 1f), Vector2.One * num, color3, 2f, Color.Black);
                    num2 += num4;
                }

                if (alpha == 1f && valid)
                {
                    // Draw next rank
                    float yoffset = 38f;
                    if (Celeste.Settings.Instance.SpeedrunClock == SpeedrunType.File)
                    {
                        yoffset += 38f;
                    }
                    bg.Draw(position + new Vector2(250, yoffset - 38f), Vector2.Zero, Color.White, 1.2f);
                    bg.Draw(position + new Vector2(-50, yoffset - 38f), Vector2.Zero, Color.White, 1.2f);
                    yoffset += 8f;
                    //yoffset = 8f;
                    //if (Celeste.Settings.Instance.SpeedrunClock == SpeedrunType.File)
                    //{
                    //    yoffset = yoffset - 4 + 26.4000015f;
                    //}
                    string tierString;
                    if (finished)
                    {
                        tierString = $"Rank: {RankToString(curRank)}";
                        if ((Engine.Scene as Level).Session.GetFlag("timeTrialsNewBest"))
                        {
                            tierString += " NEW BEST!";
                        }
                    }
                    else
                    {
                        if (targetTime != -1)
                        {
                            if (BerryTime > 0)
                            {
                                tierString = $"Next: {RankToString(curRank)}: {FormatTime(targetTime)}+{FormatTime(BerryTime)}";
                            }
                            else
                            {
                                tierString = $"Next: {RankToString(curRank)}: {FormatTime(targetTime)}";
                            }

                        }
                        else
                        {
                            tierString = $"Next: {RankToString(curRank)}";
                        }
                    }

                    num2 = position.X;
                    num3 = position.Y;
                    foreach (char c in tierString)
                    {
                        if (c == '.')
                        {
                            num = scale * 0.7f;
                            num3 -= 5f * scale;
                        }
                        Color color3 = (c == ':' || c == '.' || num < scale) ? color2 : color;
                        float num4 = (((c == ':' || c == '.') ? spacerWidth : numberWidth) + 4f) * num;
                        font.DrawOutline(fontFaceSize, c.ToString(), new Vector2(num2 + num4 / 2f, num3 + yoffset), new Vector2(0.5f, 1f), Vector2.One * num, color3, 2f, Color.Black);
                        num2 += num4;
                    }
                }


            }
            else
            {
                orig(position, timeString, scale, valid, finished, bestTime, alpha);
            }
        }

        public string RankToString(int Rank)
        {
            switch (Rank)
            {
                case 3:
                    return "Platinum";
                case 2:
                    return "Gold";
                case 1:
                    return "Silver";
                case 0:
                    return "Bronze";
                default:
                    return "None";
            }
        }
        private void SpeedrunTimerDisplay_Render(On.Celeste.SpeedrunTimerDisplay.orig_Render orig, SpeedrunTimerDisplay self)
        {
            
        }

        // Optional, initialize anything after Celeste has initialized itself properly.
        public override void Initialize()
        {
        }

        // Unload the entirety of your mod's content, remove any event listeners and undo all hooks.
        public override void Unload()
        {

        }

    }
}
